/*-
 * Copyright (c) 2013 Chris Rees <crees@FreeBSD.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <cfuconf.h>
#include <err.h>
#include <errno.h>
#include <libutil.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "keyboard-daemon.h"

static cfuconf_t *conf;
static struct pidfh *pfh;
struct st {
	char *text;
	int size;
};
static struct st key;
static FILE *dev;

/*
 * Use libcfu to read the config file, map scan codes to commands
 * and execute them
 */
int
main(int argc, char *argv[])
{
	pid_t fpid;
	int ch, dopt, i, vopt;
	char *buf, *configfile, *device, *pidfile;

	configfile = pidfile = NULL;
	pfh = NULL;
	dopt = vopt = 0;

	while ((ch = getopt(argc, argv, "Vdf:p:v")) != -1) {
		switch (ch) {
		case 'V':
			printf("keyboard-daemon version %s\n", VERSION);
			exit(0);
			break;
		case 'd':
			dopt = 1;
			break;
		case 'f':
			/* CONFFILE macro is used to get the config file */
			configfile = optarg;
			break;
		case 'p':
			pidfile = optarg;
			break;
		case 'v':
			vopt = 1;
			break;
		default:
			usage();
			break;
		}
	}

	argc -= optind;
	argv += optind;

	/* Only one remaining argument permitted; taken as device name */
	if (argc != 1)
		usage();

	device = argv[0];

	if (cfuconf_parse_file(CONFFILE, &conf, &buf) != 0) {
		if (buf != NULL) {
			warnx("While reading config file \"%s\": %s\n", 
			    CONFFILE, buf);
			free(buf);
			exit(1);
		} else {
			errx(1, "Something unknown failed while opening the config file");
		}
	}

	key.size = 1;
	for (i = 0; i < sizeof(int) * 8; i++)
		key.size *= 2;
	key.text = malloc(key.size);
	snprintf(key.text, key.size - 1, "%d", 10);

	cfuconf_get_directive_one_arg(conf, key.text, &buf);
	
	if (dopt == 1) {
		if (pidfile != NULL) {
			pfh = pidfile_open(pidfile, 0600, &fpid);
		}
		if (pfh == NULL) {
			if (errno == EEXIST)
				errx(EXIT_FAILURE,
				    "keyboard-daemon already running, pid: %d.",
				    fpid);
			err(EXIT_FAILURE, "Cannot open pidfile");
		}

		if (daemon(0, 0) == -1) {
			warn("Cannot daemonize");
			pidfile_remove(pfh);
			exit(EXIT_FAILURE);
		}

		pidfile_write(pfh);
	}

	if ((dev = fopen(device, "r")) == NULL)
		err(1, "%s", device);

	signal(SIGTERM, &cleanup);

	while ((ch = getc(dev)) != EOF) {
		snprintf(key.text, key.size - 1, "%x", ch);
		if (cfuconf_get_directive_one_arg(conf, key.text, &buf) == 0) {
			/*
			 * The input here is trusted; only root should
			 * be able to write to this config file
			 */
			fpid = fork();
			if (fpid == 0) {
				system(buf);
				exit(0);
			} else if (fpid == -1)
				err(1, "forking failed");
		} else if (vopt) {
			warnx("Unrecognised scan code: %s", key.text);
		}
	}
	
	exit(0);
}

static void
cleanup(int sig)
{

	pidfile_remove(pfh);
	fclose(dev);
	cfuconf_destroy(conf);
	free(key.text);
}

static void
usage(void)
{

	fprintf(stderr, "usage: [-Vdv] [-f conffile] [-p pidfile] device\n");
	exit(1);
}
